package com.nokia.siteinfo.task;

import com.nokia.common.gpload.GploadUtil;
import com.nokia.common.gpload.entity.GploadResult;
import com.xxl.job.core.context.XxlJobHelper;
import com.xxl.job.core.handler.annotation.XxlJob;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Component
@Slf4j
public class UpdateTask {
    @Value("${gpload.shell.path}")
    private String gploadShellPath;

    private final JdbcTemplate inputJdbcTemplate;
    private final JdbcTemplate outputJdbcTemplate;

    public UpdateTask(JdbcTemplate inputJdbcTemplate, JdbcTemplate outputJdbcTemplate) {
        this.inputJdbcTemplate = inputJdbcTemplate;
        this.outputJdbcTemplate = outputJdbcTemplate;
    }

//    0 0 1 ? * 2
    @XxlJob("siteInfoJobHandler")
    public void siteInfoJobHandler() {
        try {
            Map<Object, Object> siteLevelMap = getSiteLevelMap();
            Map<String, Map<String, Object>> four = new HashMap<>();
            Map<String, Map<String, Object>> five = new HashMap<>();
            update4g(four);
            update5g(five);
            writeCsvs(four, five, siteLevelMap);
            backup("customer_service.cfg_p_netconf_std");
            gpload("customer_service.cfg_p_netconf_std");
            backup("customer_service.cfg_cell_info");
            gpload("customer_service.cfg_cell_info");
        } catch (Exception e) {
            log.error("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.log("发生异常了: {}", e.getMessage(), e);
            XxlJobHelper.handleFail(e.getMessage());
        }
    }

    private Map<Object, Object> getSiteLevelMap() {
        Map<Object, Object> siteLevelMap = new HashMap<>();
        String sql = "select * from cfm.cfg_0_4g_sitelevel";
        List<Map<String, Object>> l1 = outputJdbcTemplate.queryForList(sql);
        log.info("cfg_0_4g_sitelevel: {}", l1.size());
        l1.forEach(t -> siteLevelMap.put(t.get("site_id"), t.get("area3")));
        sql = "select * from cfm.cfg_0_5g_sitelevel";
        List<Map<String, Object>> l2 = outputJdbcTemplate.queryForList(sql);
        log.info("cfg_0_5g_sitelevel: {}", l2.size());
        l2.forEach(t -> siteLevelMap.put(t.get("site_id"), t.get("area3")));
        log.info("siteLevelMap: {}", siteLevelMap.size());
        return siteLevelMap;
    }

    private void update4g(Map<String, Map<String, Object>> map) throws IOException {
        // 查询o2p.cfg_0_4g_siteinfo
        String sql = "select * from o2p.cfg_0_4g_siteinfo";
        List<Map<String, Object>> list = inputJdbcTemplate.queryForList(sql);
        // 去重
        list.forEach(t -> map.put((String) t.get("cgi"), t));
        log.info("cfg_0_4g_siteinfo: {} -> {}", list.size(), map.size());
        XxlJobHelper.log("cfg_0_4g_siteinfo: {} -> {}", list.size(), map.size());
        String table = "cfm.cfg_0_4g_siteinfo";
        writeCsv(map, table);
        backup(table);
        // o2p.cfg_0_4g_siteinfo -> cfm.cfg_0_4g_siteinfo
        gpload(table);
        // 查询cfm.cfg_0_4g_siteinfo_gx
        sql = "select * from cfm.cfg_0_4g_siteinfo_gx";
        list = outputJdbcTemplate.queryForList(sql);
        // 添加新元素
        list.forEach(t -> map.putIfAbsent((String) t.get("cgi"), t));
        log.info("cfg_0_4g_siteinfo_gx: {} -> {}", list.size(), map.size());
        XxlJobHelper.log("cfg_0_4g_siteinfo_gx: {} -> {}", list.size(), map.size());
    }

    private void update5g(Map<String, Map<String, Object>> map) throws IOException {
        // 查询o2p.cfg_0_5g_siteinfo
        String sql = "select * from o2p.cfg_0_5g_siteinfo";
        List<Map<String, Object>> list = inputJdbcTemplate.queryForList(sql);
        // 去重
        list.forEach(t -> map.put((String) t.get("cgisai"), t));
        log.info("cfg_0_5g_siteinfo: {} -> {}", list.size(), map.size());
        XxlJobHelper.log("cfg_0_5g_siteinfo: {} -> {}", list.size(), map.size());
        String table = "cfm.cfg_0_5g_siteinfo";
        writeCsv(map, table);
        backup(table);
        // o2p.cfg_0_5g_siteinfo -> cfm.cfg_0_5g_siteinfo
        gpload(table);
        // 查询cfm.cfg_0_5g_siteinfo_gx
        sql = "select * from cfm.cfg_0_5g_siteinfo_gx";
        list = outputJdbcTemplate.queryForList(sql);
        // 添加新元素
        list.forEach(t -> map.putIfAbsent((String) t.get("cgisai"), t));
        log.info("cfg_0_5g_siteinfo_gx: {} -> {}", list.size(), map.size());
        XxlJobHelper.log("cfg_0_5g_siteinfo_gx: {} -> {}", list.size(), map.size());
    }

    /**
     * 将cfg_0_4g_siteinfo、cfg_0_5g_siteinfo转成cfg_p_netconf_std、cfg_cell_info写入csv文件
     */
    private void writeCsvs(Map<String, Map<String, Object>> four, Map<String, Map<String, Object>> five,
                           Map<Object, Object> siteLevelMap) throws IOException {
        try (OutputStreamWriter osw1 = new OutputStreamWriter(
                Files.newOutputStream(Paths.get("customer_service.cfg_p_netconf_std.csv")),
                StandardCharsets.UTF_8);
             CSVPrinter printer1 = new CSVPrinter(osw1, CSVFormat.DEFAULT);
             OutputStreamWriter osw2 = new OutputStreamWriter(
                     Files.newOutputStream(Paths.get("customer_service.cfg_cell_info.csv")),
                     StandardCharsets.UTF_8);
             CSVPrinter printer2 = new CSVPrinter(osw2, CSVFormat.DEFAULT);) {
            Map<Object, Map<String, Object>> map = new HashMap<>();
            fourGCfgPNetconfStd(four, map, siteLevelMap);
            fiveGCfgPNetconfStd(five, map, siteLevelMap);
            log.info("cfg_p_netconf_std: {}", map.size());
            XxlJobHelper.log("cfg_p_netconf_std: {}", map.size());
            printCfgPNetconfStd(printer1, map);
            printCfgCellInfo(printer2, map);
        }
    }

    /**
     * cfg_0_5g_siteinfo -> cfg_p_netconf_std
     */
    private void fiveGCfgPNetconfStd(Map<String, Map<String, Object>> five, Map<Object, Map<String, Object>> map, Map<Object, Object> siteLevelMap) {
        for (Map<String, Object> t : five.values()) {
            Map<String, Object> m = new HashMap<>();
            BigDecimal gnbid = (BigDecimal) t.get("gnbid");
            BigDecimal cell_id = (BigDecimal) t.get("cell_id");
            BigDecimal cellId = gnbid.multiply(new BigDecimal(4096)).add(cell_id);
            BigDecimal id = cellId.add(new BigDecimal(4600100000000000L));
            m.put("id", "");
            m.put("city_id", t.get("city_code"));
            m.put("city_name", t.get("city_name"));
            m.put("area_id", t.get("district_name"));
            m.put("core_ne_name", "");
            m.put("radio_ne_name", "");
            m.put("site_name", t.get("station_name"));
            m.put("site_id", gnbid);
            m.put("cell_no", cell_id);
            m.put("cell_name", t.get("cell_name"));
            m.put("cell_id", id);
            m.put("ecgi", cellId);
            m.put("gnb_cu_id", "");
            m.put("gnb_du_id", "");
            m.put("gsm_cell_id", "");
            m.put("network_type", 5);
            m.put("type_5g", "");
            m.put("tac", "");
            m.put("cover_type", "宏站".equals(t.get("station_type")) ? "室外" : "室内");
            m.put("nodeb_type", "");
            m.put("rac", "");
            m.put("lac", "");
            m.put("grid", "");
            m.put("areatype", t.get("scene"));
            m.put("sub_areatype", "");
            m.put("area3", siteLevelMap.get(gnbid.toString()));
            m.put("longitude", t.get("lon"));
            m.put("latitude", t.get("lat"));
            m.put("vendor", t.get("vender"));
            m.put("height", t.get("height"));
            m.put("downtilt", ((BigDecimal) t.get("m_downtilt")).add((BigDecimal) t.get("e_downtilt")));
            m.put("azimuth", t.get("direction"));
            m.put("fr", t.get("down_freq"));
            m.put("scramber", "");
            m.put("province_id", t.get("河北省"));
            m.put("bsbusip", "");
            m.put("bsbusip_1", "");
            m.put("isanchor", "");
            m.put("operator", t.get("construction"));
            m.put("isshare", "");
            m.put("slice_type", "");
            map.put(id, m);
        }
    }

    /**
     * cfg_0_4g_siteinfo -> cfg_p_netconf_std
     */
    private void fourGCfgPNetconfStd(Map<String, Map<String, Object>> four, Map<Object, Map<String, Object>> map, Map<Object, Object> siteLevelMap) {
        for (Map<String, Object> t : four.values()) {
            Map<String, Object> m = new HashMap<>();
            BigDecimal enbid = (BigDecimal) t.get("enbid");
            BigDecimal cell_id = (BigDecimal) t.get("cell_id");
            BigDecimal cellId = enbid.multiply(new BigDecimal(256)).add(cell_id);
            BigDecimal id = cellId.add(new BigDecimal(46001000000000L));
            m.put("id", "");
            m.put("city_id", t.get("city_code"));
            m.put("city_name", t.get("city_name"));
            m.put("area_id", t.get("district_name"));
            m.put("core_ne_name", "");
            m.put("radio_ne_name", "");
            m.put("site_name", t.get("bbu_name"));
            m.put("site_id", enbid);
            m.put("cell_no", cell_id);
            m.put("cell_name", t.get("cell_name"));
            m.put("cell_id", id);
            m.put("ecgi", cellId);
            m.put("gnb_cu_id", "");
            m.put("gnb_du_id", "");
            m.put("gsm_cell_id", "");
            m.put("network_type", 4);
            m.put("type_5g", "");
            m.put("tac", "");
            m.put("cover_type", "宏站".equals(t.get("station_type")) ? "室外" : "室内");
            m.put("nodeb_type", "");
            m.put("rac", "");
            m.put("lac", "");
            m.put("grid", "");
            m.put("areatype", t.get("scene"));
            m.put("sub_areatype", "");
            m.put("area3", siteLevelMap.get(enbid.toString()));
            m.put("longitude", t.get("lon"));
            m.put("latitude", t.get("lat"));
            m.put("vendor", t.get("vender"));
            m.put("height", t.get("height"));
            m.put("downtilt", ((BigDecimal) t.get("m_downtilt")).add((BigDecimal) t.get("e_downtilt")));
            m.put("azimuth", t.get("direction"));
            m.put("fr", t.get("down_freq"));
            m.put("scramber", "");
            m.put("province_id", t.get("河北省"));
            m.put("bsbusip", "");
            m.put("bsbusip_1", "");
            m.put("isanchor", "");
            m.put("operator", t.get("construction"));
            m.put("isshare", "");
            m.put("slice_type", "");
            map.put(id, m);
        }
    }

    private void writeCsv(Map<String, Map<String, Object>> map, String table) throws IOException {
        log.info("writeCsv: {}", table);
        XxlJobHelper.log("writeCsv: {}", table);
        try (OutputStreamWriter osw = new OutputStreamWriter(Files.newOutputStream(Paths.get(table + ".csv")),
                StandardCharsets.UTF_8);
             CSVPrinter printer = new CSVPrinter(osw, CSVFormat.DEFAULT);) {
            for (Map<String, Object> t : map.values()) {
                printer.printRecord(t.values());
            }
        }
    }

    /**
     * 备份
     *
     * @param table 表
     */
    private void backup(String table) {
        log.info("备份: {}", table);
        XxlJobHelper.log("备份: {}", table);
        LocalDate now = LocalDate.now();
        String nowString = dateFormat(now);
        // 备份
        String sql = "create table " + table + "_bak_" + nowString + " as select * from " + table;
        outputJdbcTemplate.execute(sql);
        // 删除之前的备份
        sql = "DO $$\n" +
                "DECLARE t varchar(255);\n" +
                "BEGIN FOR t IN\n" +
                "select 'drop table ' || schemaname || '.' || tablename || ';'\n" +
                "from pg_tables\n" +
                "where schemaname || '.' || tablename like '" + table + "_bak_%' " +
                "and schemaname || '.' || tablename != '" + table + "_bak_" + nowString + "' loop EXECUTE t;\n" +
                "END LOOP;\n" +
                "END;\n" +
                "$$";
        outputJdbcTemplate.execute(sql);
        // 清空表
        sql = "truncate table " + table;
        outputJdbcTemplate.execute(sql);
    }

    /**
     * 日期格式
     *
     * @param localDate 当地日期
     * @return {@link String}
     */
    private String dateFormat(LocalDate localDate) {
        return localDate.format(DateTimeFormatter.ofPattern("MMdd"));
    }

    private void gpload(String table) throws IOException {
        String gploadCommand = "sh " + gploadShellPath + " " + table;
        GploadResult gpload = GploadUtil.gpload(gploadCommand);
        if (Boolean.TRUE.equals(gpload.getTaskStatus())) {
            log.info("gpload {} 完成: {}", table, gpload);
            XxlJobHelper.log("gpload {} 完成: {}", table, gpload);
            // 删除文件
            Files.deleteIfExists(Paths.get(table + ".csv"));
        } else {
            throw new RuntimeException("gpload " + table + " 失败: " + gpload.getMessage());
        }
    }

    /**
     * 将cfg_p_netconf_std写入csv文件
     */
    private void printCfgPNetconfStd(CSVPrinter p, Map<Object, Map<String, Object>> m) throws IOException {
        for (Map<String, Object> t : m.values()) {
            p.printRecord(
                    t.get("id"),
                    t.get("city_id"),
                    t.get("city_name"),
                    t.get("area_id"),
                    t.get("core_ne_name"),
                    t.get("radio_ne_name"),
                    t.get("site_name"),
                    t.get("site_id"),
                    t.get("cell_no"),
                    t.get("cell_name"),
                    t.get("cell_id"),
                    t.get("ecgi"),
                    t.get("gnb_cu_id"),
                    t.get("gnb_du_id"),
                    t.get("gsm_cell_id"),
                    t.get("network_type"),
                    t.get("type_5g"),
                    t.get("tac"),
                    t.get("cover_type"),
                    t.get("nodeb_type"),
                    t.get("rac"),
                    t.get("lac"),
                    t.get("grid"),
                    t.get("areatype"),
                    t.get("sub_areatype"),
                    t.get("area3"),
                    t.get("longitude"),
                    t.get("latitude"),
                    t.get("vendor"),
                    t.get("height"),
                    t.get("downtilt"),
                    t.get("azimuth"),
                    t.get("fr"),
                    t.get("scramber"),
                    t.get("province_id"),
                    t.get("bsbusip"),
                    t.get("bsbusip_1"),
                    t.get("isanchor"),
                    t.get("operator"),
                    t.get("isshare"),
                    t.get("slice_type")
            );
        }
    }

    /**
     * 将cfg_cell_info写入csv文件
     */
    private void printCfgCellInfo(CSVPrinter p, Map<Object, Map<String, Object>> m) throws IOException {
        for (Map<String, Object> t : m.values()) {
            p.printRecord(
                    t.get("cell_id"),
                    t.get("site_name"),
                    t.get("city_name"),
                    t.get("longitude"),
                    t.get("latitude"),
                    t.get("tac")
            );
        }
    }
}
